@@ -19,10 +19,9 @@ Not implemented yet:
       - placeholders are ignored
 """
 import atexit
-import ConfigParser
 import errno
+import functools
 import grp
-import hashlib
 import json
 import logging
 import os
@@ -36,6 +35,8 @@ except ImportError:
     rpmutils_present = False
 import re
 import shutil
+import six
+import six.moves.configparser as ConfigParser
 import subprocess
 import tempfile
 
@@ -49,7 +50,7 @@ LOG = logging.getLogger(__name__)
 
 
 def to_boolean(b):
-    val = b.lower().strip() if isinstance(b, basestring) else b
+    val = b.lower().strip() if isinstance(b, six.string_types) else b
     return val in [True, 'true', 'yes', '1', 1]
 
 
@@ -188,8 +189,10 @@ class CommandRunner(object):
         self._stdout = output[0]
         self._stderr = output[1]
         if self._status:
-            LOG.debug("Return code of %d after executing: '%s'" % (
-                self._status, cmd))
+            LOG.debug("Return code of %d after executing: '%s'\n"
+                      "stdout: '%s'\n"
+                      "stderr: '%s'" % (self._status, cmd, self._stdout,
+                                        self._stderr))
         if self._next:
             self._next.run()
         return self
@@ -243,7 +246,7 @@ class RpmHelper(object):
                         e.g., ['2.0', '2.2', '2.2-1.fc16', '2.2.22-1.fc16']
         """
         if versions:
-            if isinstance(versions, basestring):
+            if isinstance(versions, six.string_types):
                 return versions
             versions = sorted(versions, rpmutils.compareVerOnly,
                               reverse=True)
@@ -290,8 +293,36 @@ class RpmHelper(object):
         return command.status == 0
 
     @classmethod
-    def install(cls, packages, rpms=True):
-        """Installs (or upgrades) a set of packages via RPM or via Yum.
+    def dnf_package_available(cls, pkg):
+        """Indicates whether pkg is available via dnf.
+
+        Arguments:
+            pkg -- A package name (with optional version and release spec).
+                   e.g., httpd
+                   e.g., httpd-2.2.22
+                   e.g., httpd-2.2.22-1.fc21
+        """
+        cmd_str = "dnf -y --showduplicates list available %s" % pkg
+        command = CommandRunner(cmd_str).run()
+        return command.status == 0
+
+    @classmethod
+    def zypper_package_available(cls, pkg):
+        """Indicates whether pkg is available via zypper.
+
+        Arguments:
+            pkg -- A package name (with optional version and release spec).
+                   e.g., httpd
+                   e.g., httpd-2.2.22
+                   e.g., httpd-2.2.22-1.fc16
+        """
+        cmd_str = "zypper -n --no-refresh search %s" % pkg
+        command = CommandRunner(cmd_str).run()
+        return command.status == 0
+
+    @classmethod
+    def install(cls, packages, rpms=True, zypper=False, dnf=False):
+        """Installs (or upgrades) packages via RPM, yum, dnf, or zypper.
 
         Arguments:
             packages -- a list of packages to install
@@ -305,11 +336,25 @@ class RpmHelper(object):
                           - pkg name with version spec (httpd-2.2.22), or
                           - pkg name with version-release spec
                             (httpd-2.2.22-1.fc16)
+            zypper   -- if True:
+                        * overrides use of yum, use zypper instead
+            dnf      -- if True:
+                        * overrides use of yum, use dnf instead
+                        * packages must be in same format as yum pkg list
         """
         if rpms:
             cmd = "rpm -U --force --nosignature "
             cmd += " ".join(packages)
             LOG.info("Installing packages: %s" % cmd)
+        elif zypper:
+            cmd = "zypper -n install "
+            cmd += " ".join(packages)
+            LOG.info("Installing packages: %s" % cmd)
+        elif dnf:
+            # use dnf --best to upgrade outdated-but-installed packages
+            cmd = "dnf -y --best install "
+            cmd += " ".join(packages)
+            LOG.info("Installing packages: %s" % cmd)
         else:
             cmd = "yum -y install "
             cmd += " ".join(packages)
@@ -319,8 +364,8 @@ class RpmHelper(object):
             LOG.warn("Failed to install packages: %s" % cmd)
 
     @classmethod
-    def downgrade(cls, packages, rpms=True):
-        """Downgrades a set of packages via RPM or via Yum.
+    def downgrade(cls, packages, rpms=True, zypper=False, dnf=False):
+        """Downgrades a set of packages via RPM, yum, dnf, or zypper.
 
         Arguments:
             packages -- a list of packages to downgrade
@@ -333,9 +378,25 @@ class RpmHelper(object):
                           - pkg name with version spec (httpd-2.2.22), or
                           - pkg name with version-release spec
                             (httpd-2.2.22-1.fc16)
+            dnf     -- if True:
+                       * Use dnf instead of RPM/yum
         """
         if rpms:
             cls.install(packages)
+        elif zypper:
+            cmd = "zypper -n install --oldpackage "
+            cmd += " ".join(packages)
+            LOG.info("Downgrading packages: %s", cmd)
+            command = CommandRunner(cmd).run()
+            if command.status:
+                LOG.warn("Failed to downgrade packages: %s" % cmd)
+        elif dnf:
+            cmd = "dnf -y downgrade "
+            cmd += " ".join(packages)
+            LOG.info("Downgrading packages: %s", cmd)
+            command = CommandRunner(cmd).run()
+            if command.status:
+                LOG.warn("Failed to downgrade packages: %s" % cmd)
         else:
             cmd = "yum -y downgrade "
             cmd += " ".join(packages)
@@ -348,7 +409,7 @@ class RpmHelper(object):
 class PackagesHandler(object):
     _packages = {}
 
-    _package_order = ["dpkg", "rpm", "apt", "yum"]
+    _package_order = ["dpkg", "rpm", "apt", "yum", "dnf"]
 
     @staticmethod
     def _pkgsort(pkg1, pkg2):
@@ -373,7 +434,7 @@ class PackagesHandler(object):
         # -b == local & remote install
         # -y == install deps
         opts = '-b -y'
-        for pkg_name, versions in packages.iteritems():
+        for pkg_name, versions in packages.items():
             if len(versions) > 0:
                 cmd_str = 'gem install %s --version %s %s' % (opts,
                                                               versions[0],
@@ -385,11 +446,11 @@ class PackagesHandler(object):
     def _handle_python_packages(self, packages):
         """very basic support for easy_install."""
         # TODO(asalkeld) support versions
-        for pkg_name, versions in packages.iteritems():
+        for pkg_name, versions in packages.items():
             cmd_str = 'easy_install %s' % (pkg_name)
             CommandRunner(cmd_str).run()
 
-    def _handle_yum_packages(self, packages):
+    def _handle_zypper_packages(self, packages):
         """Handle installation, upgrade, or downgrade of packages via yum.
 
         Arguments:
@@ -412,7 +473,108 @@ class PackagesHandler(object):
         # collect pkgs for batch processing at end
         installs = []
         downgrades = []
-        for pkg_name, versions in packages.iteritems():
+        for pkg_name, versions in packages.items():
+            ver = RpmHelper.newest_rpm_version(versions)
+            pkg = "%s-%s" % (pkg_name, ver) if ver else pkg_name
+            if RpmHelper.rpm_package_installed(pkg):
+                # FIXME:print non-error, but skipping pkg
+                pass
+            elif not RpmHelper.zypper_package_available(pkg):
+                LOG.warn("Skipping package '%s' - unavailable via zypper", pkg)
+            elif not ver:
+                installs.append(pkg)
+            else:
+                current_ver = RpmHelper.rpm_package_version(pkg)
+                rc = RpmHelper.compare_rpm_versions(current_ver, ver)
+                if rc < 0:
+                    installs.append(pkg)
+                elif rc > 0:
+                    downgrades.append(pkg)
+        if installs:
+            RpmHelper.install(installs, rpms=False, zypper=True)
+        if downgrades:
+            RpmHelper.downgrade(downgrades, zypper=True)
+
+    def _handle_dnf_packages(self, packages):
+        """Handle installation, upgrade, or downgrade of packages via dnf.
+
+        Arguments:
+        packages -- a package entries map of the form:
+                      "pkg_name" : "version",
+                      "pkg_name" : ["v1", "v2"],
+                      "pkg_name" : []
+
+        For each package entry:
+          * if no version is supplied and the package is already installed, do
+            nothing
+          * if no version is supplied and the package is _not_ already
+            installed, install it
+          * if a version string is supplied, and the package is already
+            installed, determine whether to downgrade or upgrade (or do nothing
+            if version matches installed package)
+          * if a version array is supplied, choose the highest version from the
+            array and follow same logic for version string above
+        """
+        # collect pkgs for batch processing at end
+        installs = []
+        downgrades = []
+        for pkg_name, versions in packages.items():
+            ver = RpmHelper.newest_rpm_version(versions)
+            pkg = "%s-%s" % (pkg_name, ver) if ver else pkg_name
+            if RpmHelper.rpm_package_installed(pkg):
+                # FIXME:print non-error, but skipping pkg
+                pass
+            elif not RpmHelper.dnf_package_available(pkg):
+                LOG.warn("Skipping package '%s'. Not available via yum" % pkg)
+            elif not ver:
+                installs.append(pkg)
+            else:
+                current_ver = RpmHelper.rpm_package_version(pkg)
+                rc = RpmHelper.compare_rpm_versions(current_ver, ver)
+                if rc < 0:
+                    installs.append(pkg)
+                elif rc > 0:
+                    downgrades.append(pkg)
+        if installs:
+            RpmHelper.install(installs, rpms=False, dnf=True)
+        if downgrades:
+            RpmHelper.downgrade(downgrades, rpms=False, dnf=True)
+
+    def _handle_yum_packages(self, packages):
+        """Handle installation, upgrade, or downgrade of packages via yum.
+
+        Arguments:
+        packages -- a package entries map of the form:
+                      "pkg_name" : "version",
+                      "pkg_name" : ["v1", "v2"],
+                      "pkg_name" : []
+
+        For each package entry:
+          * if no version is supplied and the package is already installed, do
+            nothing
+          * if no version is supplied and the package is _not_ already
+            installed, install it
+          * if a version string is supplied, and the package is already
+            installed, determine whether to downgrade or upgrade (or do nothing
+            if version matches installed package)
+          * if a version array is supplied, choose the highest version from the
+            array and follow same logic for version string above
+        """
+
+        cmd = CommandRunner("which yum").run()
+        if cmd.status == 1:
+            # yum not available, use DNF if available
+            self._handle_dnf_packages(packages)
+            return
+        elif cmd.status == 127:
+            # `which` command not found
+            LOG.info("`which` not found. Using yum without checking if dnf "
+                     "is available")
+
+        # collect pkgs for batch processing at end
+        installs = []
+        downgrades = []
+        for pkg_name, versions in packages.items():
             ver = RpmHelper.newest_rpm_version(versions)
             pkg = "%s-%s" % (pkg_name, ver) if ver else pkg_name
             if RpmHelper.rpm_package_installed(pkg):
@@ -429,10 +591,10 @@ class PackagesHandler(object):
                     installs.append(pkg)
                 elif rc > 0:
                     downgrades.append(pkg)
-            if installs:
-                RpmHelper.install(installs, rpms=False)
-            if downgrades:
-                RpmHelper.downgrade(downgrades)
+        if installs:
+            RpmHelper.install(installs, rpms=False)
+        if downgrades:
+            RpmHelper.downgrade(downgrades)
 
     def _handle_rpm_packages(self, packages):
         """Handle installation, upgrade, or downgrade of packages via rpm.
@@ -460,6 +622,8 @@ class PackagesHandler(object):
 
     # map of function pointers to handle different package managers
     _package_handlers = {"yum": _handle_yum_packages,
+                         "dnf": _handle_dnf_packages,
+                         "zypper": _handle_zypper_packages,
                          "rpm": _handle_rpm_packages,
                          "apt": _handle_apt_packages,
                          "rubygems": _handle_gem_packages,
@@ -480,10 +644,19 @@ class PackagesHandler(object):
           * rpm
           * apt
           * yum
+          * dnf
         """
         if not self._packages:
             return
-        packages = sorted(self._packages.iteritems(), PackagesHandler._pkgsort)
+        try:
+            packages = sorted(
+                self._packages.items(), cmp=PackagesHandler._pkgsort)
+        except TypeError:
+            # On Python 3, we have to use key instead of cmp
+            # This could also work on Python 2.7, but not on 2.6
+            packages = sorted(
+                self._packages.items(),
+                key=functools.cmp_to_key(PackagesHandler._pkgsort))
 
         for manager, package_entries in packages:
             handler = self._package_handler(manager)
@@ -500,7 +673,7 @@ class FilesHandler(object):
     def apply_files(self):
         if not self._files:
             return
-        for fdest, meta in self._files.iteritems():
+        for fdest, meta in self._files.items():
             dest = fdest.encode()
             try:
                 os.makedirs(os.path.dirname(dest))
@@ -511,16 +684,17 @@ class FilesHandler(object):
                     LOG.exception(e)
 
             if 'content' in meta:
-                if isinstance(meta['content'], basestring):
+                if isinstance(meta['content'], six.string_types):
                     f = open(dest, 'w+')
                     f.write(meta['content'])
                     f.close()
                 else:
                     f = open(dest, 'w+')
-                    f.write(json.dumps(meta['content'], indent=4))
+                    f.write(json.dumps(meta['content'], indent=4)
+                     .encode('UTF-8'))
                     f.close()
             elif 'source' in meta:
-                CommandRunner('wget -O %s %s' % (dest, meta['source'])).run()
+                CommandRunner('curl -o %s %s' % (dest, meta['source'])).run()
             else:
                 LOG.error('%s %s' % (dest, str(meta)))
                 continue
@@ -599,20 +773,20 @@ class SourcesHandler(object):
         basename = os.path.basename(url)
         stype = self._source_type(url)
         if stype == '.tgz':
-            cmd = "wget -q -O - '%s' | gunzip | tar -xvf -" % url
+            cmd = "curl -s '%s' | gunzip | tar -xvf -" % url
         elif stype == '.tbz2':
-            cmd = "wget -q -O - '%s' | bunzip2 | tar -xvf -" % url
+            cmd = "curl -s '%s' | bunzip2 | tar -xvf -" % url
         elif stype == '.zip':
             tmp = self._url_to_tmp_filename(url)
-            cmd = "wget -q -O '%s' '%s' && unzip -o '%s'" % (tmp, url, tmp)
+            cmd = "curl -s -o '%s' '%s' && unzip -o '%s'" % (tmp, url, tmp)
         elif stype == '.tar':
-            cmd = "wget -q -O - '%s' | tar -xvf -" % url
+            cmd = "curl -s '%s' | tar -xvf -" % url
         elif stype == '.gz':
             (r, ext) = self._splitext(basename)
-            cmd = "wget -q -O - '%s' | gunzip > '%s'" % (url, r)
+            cmd = "curl -s '%s' | gunzip > '%s'" % (url, r)
         elif stype == '.bz2':
             (r, ext) = self._splitext(basename)
-            cmd = "wget -q -O - '%s' | bunzip2 > '%s'" % (url, r)
+            cmd = "curl -s '%s' | bunzip2 > '%s'" % (url, r)
 
         if cmd != '':
             cmd = "mkdir -p '%s'; cd '%s'; %s" % (dest, dest, cmd)
@@ -628,7 +802,7 @@ class SourcesHandler(object):
     def apply_sources(self):
         if not self._sources:
             return
-        for dest, url in self._sources.iteritems():
+        for dest, url in self._sources.items():
             self._apply_source(dest, url)
 
 
@@ -641,37 +815,47 @@ class ServicesHandler(object):
         self.hooks = hooks
 
     def _handle_sysv_command(self, service, command):
-        service_exe = "/sbin/service"
-        enable_exe = "/sbin/chkconfig"
-        cmd = ""
-        if "enable" == command:
-            cmd = "%s %s on" % (enable_exe, service)
-        elif "disable" == command:
-            cmd = "%s %s off" % (enable_exe, service)
-        elif "start" == command:
-            cmd = "%s %s start" % (service_exe, service)
-        elif "stop" == command:
-            cmd = "%s %s stop" % (service_exe, service)
-        elif "status" == command:
-            cmd = "%s %s status" % (service_exe, service)
-        command = CommandRunner(cmd)
-        command.run()
-        return command
+        if os.path.exists("/bin/systemctl"):
+            service_exe = "/bin/systemctl"
+            service = '%s.service' % service
+            service_start = '%s start %s'
+            service_status = '%s status %s'
+            service_stop = '%s stop %s'
+        elif os.path.exists("/sbin/service"):
+            service_exe = "/sbin/service"
+            service_start = '%s %s start'
+            service_status = '%s %s status'
+            service_stop = '%s %s stop'
+        else:
+            service_exe = "/usr/sbin/service"
+            service_start = '%s %s start'
+            service_status = '%s %s status'
+            service_stop = '%s %s stop'
+
+        if os.path.exists("/bin/systemctl"):
+            enable_exe = "/bin/systemctl"
+            enable_on = '%s enable %s'
+            enable_off = '%s disable %s'
+        elif os.path.exists("/sbin/chkconfig"):
+            enable_exe = "/sbin/chkconfig"
+            enable_on = '%s %s on'
+            enable_off = '%s %s off'
+        else:
+            enable_exe = "/usr/sbin/update-rc.d"
+            enable_on = '%s %s enable'
+            enable_off = '%s %s disable'
 
-    def _handle_systemd_command(self, service, command):
-        exe = "/bin/systemctl"
         cmd = ""
-        service = '%s.service' % service
         if "enable" == command:
-            cmd = "%s enable %s" % (exe, service)
+            cmd = enable_on % (enable_exe, service)
         elif "disable" == command:
-            cmd = "%s disable %s" % (exe, service)
+            cmd = enable_off % (enable_exe, service)
         elif "start" == command:
-            cmd = "%s start %s" % (exe, service)
+            cmd = service_start % (service_exe, service)
         elif "stop" == command:
-            cmd = "%s stop %s" % (exe, service)
+            cmd = service_stop % (service_exe, service)
         elif "status" == command:
-            cmd = "%s status %s" % (exe, service)
+            cmd = service_status % (service_exe, service)
         command = CommandRunner(cmd)
         command.run()
         return command
@@ -708,22 +892,21 @@ class ServicesHandler(object):
                 if start_cmd.status != 0:
                     LOG.warning('Service %s did not start. STDERR: %s' %
                                (service, start_cmd.stderr))
-                    return
                 for h in self.hooks:
                     h.event('service.restarted', service, self.resource)
 
     def _monitor_services(self, handler, services):
-        for service, properties in services.iteritems():
+        for service, properties in services.items():
             self._monitor_service(handler, service, properties)
 
     def _initialize_services(self, handler, services):
-        for service, properties in services.iteritems():
+        for service, properties in services.items():
             self._initialize_service(handler, service, properties)
 
     # map of function pointers to various service handlers
     _service_handlers = {
         "sysvinit": _handle_sysv_command,
-        "systemd": _handle_systemd_command
+        "systemd": _handle_sysv_command
     }
 
     def _service_handler(self, manager_name):
@@ -736,7 +919,7 @@ class ServicesHandler(object):
         """Starts, stops, enables, disables services."""
         if not self._services:
             return
-        for manager, service_entries in self._services.iteritems():
+        for manager, service_entries in self._services.items():
             handler = self._service_handler(manager)
             if not handler:
                 LOG.warn("Skipping invalid service type: %s" % manager)
@@ -747,7 +930,7 @@ class ServicesHandler(object):
         """Restarts failed services, and runs hooks."""
         if not self._services:
             return
-        for manager, service_entries in self._services.iteritems():
+        for manager, service_entries in self._services.items():
             handler = self._service_handler(manager)
             if not handler:
                 LOG.warn("Skipping invalid service type: %s" % manager)
@@ -827,6 +1010,10 @@ def metadata_server_port(
         return None
 
 
+class CommandsHandlerRunError(Exception):
+    pass
+
+
 class CommandsHandler(object):
 
     def __init__(self, commands):
@@ -885,14 +1072,12 @@ class CommandsHandler(object):
         if command_status == 0:
             LOG.info("%s has been successfully executed" % command_label)
         else:
-            if "ignoreErrors" in properties:
-                if properties["ignoreErrors"] == "false":
-                    LOG.error("%s has failed. Not ignoring" % command_label)
-                else:
-                    LOG.info("%s has failed. Explicit ignoring"
-                             % command_label)
+            if "ignoreErrors" in properties and \
+               to_boolean(properties["ignoreErrors"]):
+                LOG.info("%s has failed (status=%d). Explicit ignoring"
+                         % (command_label, command_status))
             else:
-                LOG.error("%s has failed." % command_label)
+                raise CommandsHandlerRunError("%s has failed." % command_label)
 
 
 class GroupsHandler(object):
@@ -904,7 +1089,7 @@ class GroupsHandler(object):
         """Create Linux/UNIX groups and assign group IDs."""
         if not self.groups:
             return
-        for group, properties in self.groups.iteritems():
+        for group, properties in self.groups.items():
             LOG.debug("%s group is being created" % group)
             self._initialize_group(group, properties)
 
@@ -948,7 +1133,7 @@ class UsersHandler(object):
         """Create Linux/UNIX users and assign user IDs, groups and homedir."""
         if not self.users:
             return
-        for user, properties in self.users.iteritems():
+        for user, properties in self.users.items():
             LOG.debug("%s user is being created" % user)
             self._initialize_user(user, properties)
 
@@ -1065,7 +1250,7 @@ class Metadata(object):
 
         url = 'http://169.254.169.254/openstack/2012-08-10/meta_data.json'
         if not os.path.exists(cache_path):
-            CommandRunner('wget -O %s %s' % (cache_path, url)).run()
+            CommandRunner('curl -o %s %s' % (cache_path, url)).run()
         try:
             with open(cache_path) as fd:
                 try:
@@ -1076,6 +1261,14 @@ class Metadata(object):
             pass
         return None
 
+    def get_instance_id(self):
+        """Get the unique identifier for this server."""
+        instance_id = None
+        md = self.get_nova_meta()
+        if md is not None:
+            instance_id = md.get('uuid')
+        return instance_id
+
     def get_tags(self):
         """Get the tags for this server."""
         tags = {}
@@ -1096,6 +1289,11 @@ class Metadata(object):
                True -- success
               False -- error
         """
+        if self.resource is not None:
+            res_last_path = last_path + '_' + self.resource
+        else:
+            res_last_path = last_path
+
         if meta_str:
             self._data = meta_str
         else:
@@ -1115,7 +1313,7 @@ class Metadata(object):
                 # cached metadata or the logic below could re-run a stale
                 # cfn-init-data
                 fd = None
-                for filepath in [last_path, default_path]:
+                for filepath in [res_last_path, last_path, default_path]:
                     try:
                         fd = open(filepath)
                     except IOError:
@@ -1138,18 +1336,21 @@ class Metadata(object):
         else:
             self._metadata = self._data
 
-        cm = hashlib.md5(json.dumps(self._metadata))
-        current_md5 = cm.hexdigest()
-        old_md5 = None
+        last_data = ""
+        for metadata_file in [res_last_path, last_path]:
+            try:
+                with open(metadata_file) as lm:
+                    try:
+                        last_data = json.load(lm)
+                    except ValueError:
+                        pass
+                    lm.close()
+            except IOError:
+                LOG.warn("Unable to open local metadata : %s" %
+                         metadata_file)
+                continue
 
-        try:
-            with open(last_path) as lm:
-                om = hashlib.md5()
-                om.update(lm.read())
-                old_md5 = om.hexdigest()
-        except Exception:
-            pass
-        if old_md5 != current_md5:
+        if self._metadata != last_data:
             self._has_changed = True
 
         # if cache dir does not exist try to create it
@@ -1167,8 +1368,11 @@ class Metadata(object):
                                          mode='wb',
                                          delete=False) as cf:
             os.chmod(cf.name, 0o600)
-            cf.write(json.dumps(self._metadata))
+            cf.write(json.dumps(self._metadata).encode('UTF-8'))
             os.rename(cf.name, last_path)
+            cf.close()
+            if res_last_path != last_path:
+                shutil.copy(last_path, res_last_path)
 
         return True
 
